import json
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol

from .subprocess_utils import shell_output


@dataclass(frozen=True)
class Issue:
    entity_id: str
    title: str


class IssueService(Protocol):
    def setup(self) -> None:
        """Any setup needed, including installing CLI tools, etc."""
        ...

    def get_issues_assigned_to_me(self) -> Sequence[Issue]:
        ...


class GithubIssueService(IssueService):
    def setup(self) -> None:
        pass

    def get_issues_assigned_to_me(self) -> Sequence[Issue]:
        """Get issues assigned to current user on current repo"""
        my_issues_cmd = shell_output("gh issue list --assignee='@me' --json number,title")

        if my_issues_cmd is None:
            return []

        values = json.loads(my_issues_cmd)
        parsed_output = [Issue(entity_id=str(v["number"]), title=v["title"]) for v in values]

        return parsed_output



from collections.abc import Sequence
from typing import Protocol

import questionary
from rich.console import Console

from .issue_service import Issue

console = Console()


class IssuePresenter(Protocol):
    def select_issue_dialog(self, issues: Sequence[Issue]) -> Issue:
        ...


class DefaultIssuePresenter(IssuePresenter):
    def select_issue_dialog(self, issues: Sequence[Issue]) -> Issue:
        selected_issue_title = questionary.select(
            "I found these issues for you. Which one do you want to work on?",
            choices=[issue.title for issue in issues],
            qmark="",
        ).ask()

        return next(issue for issue in issues if issue.title == selected_issue_title)



import typer
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm

from .registry import issue_services, presenters, stackers

app = typer.Typer()

issue_service = issue_services["Github"]()
issue_presenter = presenters["Default"]()
stacker = stackers["Graphite"]()


def look_for_new_issues() -> bool:
    retry = Confirm.ask(
        ":palm_tree: No issues assigned to you in this repository. Do you want to retry?",
        default=True,
    )
    return retry


@app.command()
def new():
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
    ) as progress:
        progress.add_task("Getting issues assigned to you", start=True)
        my_issues = issue_service.get_issues_assigned_to_me()

    if not my_issues:
        if look_for_new_issues():
            next()
        return

    selected_issue = issue_presenter.select_issue_dialog(my_issues)
    stacker.create_stack_from_trunk(selected_issue)


@app.command()
def next():  # noqa: A001 [Shadowing python built-in]
    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True
    ) as progress:
        progress.add_task("Getting issues assigned to you", start=True)
        my_issues = issue_service.get_issues_assigned_to_me()

    if not my_issues:
        if look_for_new_issues():
            next()
        return

    selected_issue = issue_presenter.select_issue_dialog(my_issues)
    stacker.add_to_stack(selected_issue)


@app.command()
def submit(automerge: bool = False):
    stacker.submit_stack(automerge=automerge)
    print(":rocket: [bold green]Stack submitted![/bold green]")
    stacker.status()


if __name__ == "__main__":
    app()



from .issue_presenter import DefaultIssuePresenter, IssuePresenter
from .issue_service import GithubIssueService, IssueService
from .stacker import Graphite, Stacker

issue_services: dict[str, type[IssueService]] = {"Github": GithubIssueService}
stackers: dict[str, type[Stacker]] = {"Graphite": Graphite}
presenters: dict[str, type[IssuePresenter]] = {"Default": DefaultIssuePresenter}



from typing import Protocol

from .issue_service import Issue
from .subprocess_utils import interactive_cmd


def sanitise_issue_title(issue_title: str) -> str:
    char2replacement = {
        " ": "-",
        ":": "/",
        ",": "",
        "'": "",
        '"': "",
        "(": "",
        ")": "",
        "[": "",
        "": "",
        "`": "",
        ">": "",
        "<": "",
        "=": "",
    }

    for character, replacement in char2replacement.items():
        issue_title = issue_title.replace(character, replacement)

    issue_title = issue_title.replace("--", "-")
    return issue_title


class Stacker(Protocol):
    def create_stack_from_trunk(self, issue: Issue):
        ...

    def add_to_stack(self, issue: Issue):
        ...

    def submit_stack(self, automerge: bool):
        ...

    def status(self):
        ...


class Graphite(Stacker):
    def _sync(self):
        interactive_cmd("gt sync --force --show-delete-progress")

    def create_stack_from_trunk(self, issue: Issue):
        self._sync()
        interactive_cmd("git checkout main")
        interactive_cmd("git pull")
        self.add_to_stack(issue)

    def add_to_stack(self, issue: Issue):
        self._sync()
        sanitised_title = sanitise_issue_title(issue.title)
        branch_title = f"{issue.entity_id}-{sanitised_title}"
        first_commit_str = f"'{issue.title}\n\nFixes #{issue.entity_id}'"
        interactive_cmd(f"gt create {branch_title} --all -m {first_commit_str}")
        interactive_cmd(f"git commit --allow-empty -m {first_commit_str}")

    def submit_stack(self, automerge: bool):
        self._sync()
        submit_command = "gt submit --no-edit --publish"

        if automerge:
            submit_command += " --merge-when-ready"

        interactive_cmd(submit_command)

    def status(self):
        interactive_cmd("gt log short --reverse")
